from itertools import product
from collections import defaultdict

import numpy as np
from joblib import Parallel, delayed

from ..base import FusionFit
from ._dfmc import dfmc


__all__ = ['Dfmc']


def parallel_dfmc_wrapper(**params):
    return dfmc(**params)


class Dfmc(FusionFit):
    """Data fusion by matrix completion.

    Attributes
    ---------
    fusion_graph :
    max_iter :
    init_type :
    n_run :
    stopping :
    stopping_system :
    verbose :
    compute_err :
    callback :
    random_state :
    n_jobs :

    Parameters
    ----------
    max_iter :
    init_type :
    n_run :
    stopping :
    stopping_system :
    verbose :
    compute_err :
    callback :
    random_state :
    n_jobs :
    """
    def __init__(self, max_iter=100, init_type='random_c', n_run=1,
                 stopping=None, stopping_system=None, verbose=0,
                 compute_err=False, callback=None,
                 random_state=None, n_jobs=1):
        super(Dfmc, self).__init__()
        self._set_params(vars())

    def fuse(self, fusion_graph):
        """Run data fusion completion algorithm.

        Parameters
        ----------
        fusion_graph :
        """
        self.fusion_graph = fusion_graph
        if not isinstance(self.random_state, np.random.RandomState):
            self.random_state = np.random.RandomState(self.random_state)

        object_types = set([ot for ot in self.fusion_graph.object_types])
        object_type2rank = {ot: int(ot.rank) for ot in self.fusion_graph.object_types}

        R, T, M = {}, {}, {}
        for row_type, col_type in product(self.fusion_graph.object_types, repeat=2):
            for relation in self.fusion_graph.get_relations(row_type, col_type):
                filled_data = relation.filled()
                if relation.preprocessor:
                    preprocessed_data = relation.preprocessor(filled_data)
                else:
                    preprocessed_data = filled_data
                if np.ma.is_masked(preprocessed_data):
                    data = preprocessed_data.data
                    mask = preprocessed_data.mask
                else:
                    data = preprocessed_data
                    mask = None
                if relation.row_type != relation.col_type:
                    R[relation.row_type, relation.col_type] = R.get((
                        relation.row_type, relation.col_type), [])
                    R[relation.row_type, relation.col_type].append(data)

                    M[relation.row_type, relation.col_type] = M.get((
                        relation.row_type, relation.col_type), [])
                    M[relation.row_type, relation.col_type].append(mask)
                else:
                    T[relation.row_type, relation.col_type] = T.get((
                        relation.row_type, relation.col_type), [])
                    T[relation.row_type, relation.col_type].append(data)

        parallelizer = Parallel(n_jobs=self.n_jobs, max_nbytes=1e3, verbose=self.verbose)
        task_iter = (delayed(parallel_dfmc_wrapper)(
            R=R, M=M, Theta=T, obj_types=object_types,
            obj_type2rank=object_type2rank, max_iter=self.max_iter, init_type=self.init_type,
            stopping=self.stopping, stopping_system=self.stopping_system, verbose=self.verbose,
            compute_err=self.compute_err, callback=self.callback, random_state=self.random_state,
            n_jobs=self.n_jobs)
                     for _ in range(self.n_run))
        entries = parallelizer(task_iter)

        self.factors_ = defaultdict(list)
        self.backbones_ = defaultdict(list)
        for G, S in entries:
            for (object_type, _), factor in G.items():
                self.factors_[object_type].append(factor)

            for (row_type, col_type), backbones in S.items():
                for i, relation in enumerate(self.fusion_graph.get_relations(row_type, col_type)):
                    self.backbones_[relation].append(backbones[i])
        return self



#! /usr/bin/env python
descr = """A Python module for data fusion built on top of factorized models."""

from setuptools import setup, find_packages
import subprocess
import os
import imp

DISTNAME = 'scikit-fusion'
DESCRIPTION = 'A Python module for data fusion built on top of factorized models.'
LONG_DESCRIPTION = open('README.rst').read()
MAINTAINER = 'Marinka Zitnik'
MAINTAINER_EMAIL = 'marinka.zitnik@fri.uni-lj.si'
URL = 'http://github.com/marinkaz/scikit-fusion'
LICENSE = 'GPLv3'
VERSION = '0.2.1'
ISRELEASED = False

INSTALL_REQUIRES = (
    'numpy>=1.7.0',
    'scipy>=0.12.0',
    'joblib>=0.8.4',
)

# Return the git revision as a string
def git_version():
    """Return the git revision as a string.

    Copied from numpy setup.py
    """
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout = subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"
    return GIT_REVISION


def write_version_py(filename='skfusion/version.py'):
    # Copied from numpy setup.py
    cnt = """
# THIS FILE IS GENERATED FROM SCIKIT-FUSION SETUP.PY
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s

if not release:
    version = full_version
    short_version += ".dev"
"""
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION = git_version()
    elif os.path.exists('skfusion/version.py'):
        # must be a source distribution, use existing version file
        version = imp.load_source("skfusion.version", "skfusion/version.py")
        GIT_REVISION = version.git_revision
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += '.dev0+' + GIT_REVISION[:7]

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()


def setup_package():
    write_version_py()
    setup(name=DISTNAME,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          long_description=LONG_DESCRIPTION,
          classifiers=['Intended Audience :: Science/Research',
                       'Intended Audience :: Developers',
                       'License :: OSI Approved',
                       'Programming Language :: Python',
                       'Topic :: Software Development',
                       'Topic :: Scientific/Engineering',
                       'Topic :: Scientific/Engineering :: Artificial Intelligence',
                       'Topic :: Scientific/Engineering :: Bio-Informatics',
                       'Programming Language :: Python :: 2',
                       'Programming Language :: Python :: 3',
                       ],
          install_requires=INSTALL_REQUIRES,
          packages=find_packages(),
          package_data={
              'skfusion.datasets': ['data/*/*']
          },
          test_suite='skfusion.tests.test_suite',
)


if __name__ == "__main__":
    setup_package()




# THIS FILE IS GENERATED FROM SCIKIT-FUSION SETUP.PY
short_version = '0.2.1'
version = '0.2.1'
full_version = '0.2.1.dev0+0159eb8'
git_revision = '0159eb86a47f277072ea4cf1d30798e2f751f7d4'
release = False

if not release:
    version = full_version
    short_version += ".dev"
